﻿/* /////////////////////////////////////////////////////////////////////////////////////////////////
SpeechRecognition by MacGruber.
Recognizes keywords or complex grammar using Windows 10 and Unity's speech recognition features.

Version 2: 2023-03-20
	Fixed null reference exception with VaM 1.22
	Fixed another silent null reference exception.

Version 1: 2021-05-15
	Initial release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Windows.Speech;
using System;
using System.Text;
using System.Collections.Generic;
using SimpleJSON;
using MVR.FileManagementSecure;

namespace MacGruber
{
    public class SpeechRecognition : MVRScript
    {
		private JSONStorableStringChooser myPhraseChooser;
		private JSONStorableBool myLoggingEnabled;
		private JSONStorableString mySetEnabled;
		private JSONStorableString mySetDisabled;
		private JSONStorableString myPhraseName;
		private JSONStorableBool myPhraseIsEnabled;
		private JSONStorableStringChooser myPhraseTypeChooser;
		private JSONStorableUrl myPhrasePath;
		private JSONStorableString myPhraseKeyword;

		private List<object> myPhraseUI = new List<object>();
		private List<Phrase> myPhrases = new List<Phrase>();

		private bool myNeedUIRefresh = false;
		private bool myNeedUIRebuild = false;

		// WHY would ANY sane person use ".xml" files with ".json" extension??!
		// Because it saves your life when saving your scene for a VAR package,
		// as VaM only tracks dependencies for a number of file types,
		// using ".json" is the least terrible choice.
		private static readonly string FILE_EXTENSION = "json";

		private static readonly string EXTRACT_PATH = "Custom/PluginData/SpeechRecognition/";
		private static readonly List<string> PHRASE_TYPE = new List<string>() { "KeywordRecognizer", "GrammarRecognizer" };
		private const int PHRASE_TYPE_KEYWORD = 0;
		private const int PHRASE_TYPE_GRAMMAR = 1;

		private static string myApplicationPath;
		private static bool myDoLogging = true;
		private static List<string> myExtractedPackages = new List<string>();

		private class Phrase
		{
			public Phrase(SpeechRecognition script, string name)
			{
				myName = name;
				Trigger = new EventTrigger(script, "On" + name);
			}

			public Phrase(SpeechRecognition script, string name, Phrase source)
			{
				Trigger = new EventTrigger(source.Trigger);
				Name = name;
				Enabled = source.Enabled;

				if (source.Path != null)
					Path = source.Path;
				else if (source.Keywords != null && source.Keywords.Count > 0)
					Keywords = new List<string>(source.Keywords);
			}

			public static int SortByNameAscending(Phrase a, Phrase b)
			{
				return a.Name.CompareTo(b.Name);
			}

			public string Name {
				get { return myName; }
				set {
					myName = value;
					Trigger.Name = "On" + value;
				}
			}

			public int Type {
				get { return myType; }
			}

			public string Path {
				get { return myPath; }
				set {
					myPath = (value != null) ? SuperController.singleton.NormalizeLoadPath(value) : null;
					myKeywords = null;
					myType = PHRASE_TYPE_GRAMMAR;
					DisposeRecognizer();

					if (myPath == null)
						return;

					string loadPath;
					if (FileManagerSecure.IsFileInPackage(myPath))
					{
						// GrammarRecognizer can't load from VAR packages, we need to create a copy of the file.
						string packageName = myPath.Substring(0, myPath.IndexOf(':'));
						if (!myExtractedPackages.Contains(packageName))
						{
							myExtractedPackages.Add(packageName);
							SuperController.LogMessage($"SpeechRecognition: Extracting Grammar XML files from package '{packageName}' to '{EXTRACT_PATH}'.");
							ExtractFilesFromPackage(packageName+":");
						}
						loadPath = myApplicationPath + EXTRACT_PATH + myPath.Replace(":/", "/");
					}
					else if (FileManagerSecure.FileExists(myPath))
					{
						loadPath = myApplicationPath + myPath;
					}
					else
					{
						SuperController.LogError($"SpeechRecognition: Failed to load Grammar phrase '{myName}'. File not found:\n{myPath}");
						return;
					}

					try {
						if (PhraseRecognitionSystem.Status != SpeechSystemStatus.Running)
							PhraseRecognitionSystem.Restart();
						myPhraseRecognizer = new GrammarRecognizer(loadPath);
						myPhraseRecognizer.OnPhraseRecognized += OnPhraseRecognized;
						StartRecognizer();
						if (myDoLogging)
							SuperController.LogMessage($"SpeechRecognition: Loaded Grammar phrase '{myName}'");
					}
					catch (Exception e)
					{
						SuperController.LogError($"SpeechRecognition: Failed to load Grammar phrase '{myName}'. File path:\n{myPath}");
					}
				}
			}

			public List<string> Keywords {
				get { return myKeywords; }
				set {
					myPath = null;
					myKeywords = value;
					myType = PHRASE_TYPE_KEYWORD;
					DisposeRecognizer();

					if (myKeywords == null || myKeywords.Count == 0)
						return;

					List<string> trimmedKeywords = new List<string>(myKeywords);
					trimmedKeywords.RemoveAll(k => string.IsNullOrEmpty(k));
					if (trimmedKeywords.Count == 0)
						return;

					try {
						if (PhraseRecognitionSystem.Status != SpeechSystemStatus.Running)
							PhraseRecognitionSystem.Restart();
						myPhraseRecognizer = new KeywordRecognizer(trimmedKeywords.ToArray());
						myPhraseRecognizer.OnPhraseRecognized += OnPhraseRecognized;
						StartRecognizer();
						if (myDoLogging)
							SuperController.LogMessage($"SpeechRecognition: Loaded Keyword phrase '{myName}'");
					}
					catch (Exception e)
					{
						SuperController.LogError($"SpeechRecognition: Failed to load Keyword phrase '{myName}'.");
					}
				}
			}

			public bool Enabled {
				get { return myEnabled; }
				set {
					myEnabled = value;
					if (myEnabled)
						StartRecognizer();
					else
						StopRecognizer();
				}
			}

			public void StartRecognizer()
			{
				if (myEnabled && myPhraseRecognizer != null && !myPhraseRecognizer.IsRunning)
					myPhraseRecognizer.Start();
			}

			public void StopRecognizer()
			{
				if (myPhraseRecognizer != null && myPhraseRecognizer.IsRunning)
					myPhraseRecognizer.Stop();
			}

			public void DisposeRecognizer()
			{
				if (myPhraseRecognizer != null)
				{
					StopRecognizer();
					myPhraseRecognizer.OnPhraseRecognized -= OnPhraseRecognized;
					myPhraseRecognizer.Dispose();
					myPhraseRecognizer = null;
				}
			}

			private void OnPhraseRecognized(PhraseRecognizedEventArgs args)
			{
				if (myDoLogging)
					SuperController.LogMessage($"SpeechRecognition: Trigger {myName} ('{args.text}', Confidence: {args.confidence})");
				Trigger.Trigger();
			}

			public EventTrigger Trigger;
			private string myName;
			private string myPath;
			private List<string> myKeywords;
			private PhraseRecognizer myPhraseRecognizer;
			private int myType = PHRASE_TYPE_KEYWORD;
			private bool myEnabled = true;
		}



        public override void Init()
		{
			if (!PhraseRecognitionSystem.isSupported)
			{
				Utils.SetupInfoText(this,
					"<color=#606060><size=40><b>SpeechRecognition</b></size>\nRecognizes keywords or complex grammar using Windows 10 and Unity's speech recognition features. This is <i>not</i> \"AI\", it does only understand phrases you teach it to understand.</color>\n\n<color=#FF0000><b>Apparently speech recognition is not supported on your machine!</b></color>",
					390.0f, true
				);
				return;
			}

			myApplicationPath = Application.dataPath;
			if (myApplicationPath.EndsWith("/VaM_Data"))
				myApplicationPath = myApplicationPath.Substring(0, myApplicationPath.Length-8);

			PhraseRecognitionSystem.OnError += OnPhaseSystemError;

			Utils.OnInitUI(CreateUIElement);

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>SpeechRecognition</b></size>\nRecognizes keywords or complex grammar using Windows 10 and Unity's speech recognition features. This is <i>not</i> \"AI\", it does only understand phrases you teach it to understand.</color>\n\n<color=#FF0000>This plugin is using your <b>system default microphone</b>, not your microphone setting in Oculus or SteamVR software!</color>",
				390.0f, true
			);

			myPhraseChooser = new JSONStorableStringChooser("Phrase", new List<string>(), "", "Phrase");
			myPhraseChooser.isStorable = myPhraseChooser.isRestorable = false;
			myPhraseChooser.displayChoices = new List<string>();
			myPhraseChooser.setCallbackFunction += UISetPhrase;
			CreateScrollablePopup(myPhraseChooser, false);

			Utils.SetupTwinButton(this, "Add Phrase", UIAddPhrase, "Duplicate Phrase", UIDuplicatePhrase, false);
			Utils.SetupButton(this, "Remove Phrase", UIRemovePhrase, false);

			myLoggingEnabled = Utils.SetupToggle(this, "Logging Enabled", true, false);
			myLoggingEnabled.setCallbackFunction += (bool v) => { myDoLogging = v; };

			Utils.SetupButton(this, "Test Trigger", UITriggerPhrase, false);

			mySetEnabled = new JSONStorableString("SetPhraseEnabled", "", SetEnabledAction);
			mySetEnabled.isStorable = mySetEnabled.isRestorable = false;
			RegisterString(mySetEnabled);
			Utils.SetupAction(this, "SetPhraseEnabledAll", () => SetEnabledAllAction(true));

			mySetDisabled = new JSONStorableString("SetPhraseDisabled", "", SetDisabledAction);
			mySetDisabled.isStorable = mySetDisabled.isRestorable = false;
			RegisterString(mySetDisabled);
			Utils.SetupAction(this, "SetPhraseDisabledAll", () => SetEnabledAllAction(false));

			Utils.SetupSpacer(this, 15, false);

			myPhraseName = new JSONStorableString("Name", "", UIRenamePhrase);
			myPhraseIsEnabled = new JSONStorableBool("Enabled", true, UISetEnabledPhrase);

			myPhraseTypeChooser = new JSONStorableStringChooser("Type", PHRASE_TYPE, PHRASE_TYPE[PHRASE_TYPE_KEYWORD], "Type");
			myPhraseTypeChooser.setCallbackFunction += UISetType;
			myPhrasePath = new JSONStorableUrl("Path", "", FILE_EXTENSION);

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void UISetPhrase(string v)
		{
			myPhraseChooser.valNoCallback = "";
			myPhraseChooser.valNoCallback = v;
			Utils.RemoveUIElements(this, myPhraseUI);

			Phrase phrase = UIGetPhrase();
			if (phrase == null)
				return;

			UIDynamicTextInfo optionHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Phrase</b></size>", false);
			myPhraseUI.Add(optionHeader);

			myPhraseName.valNoCallback = phrase.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myPhraseName, false);
			myPhraseUI.Add(nameInput);

			myPhraseIsEnabled.valNoCallback = phrase.Enabled;
			CreateToggle(myPhraseIsEnabled, false);
			myPhraseUI.Add(myPhraseIsEnabled);

			myPhraseTypeChooser.valNoCallback = PHRASE_TYPE[phrase.Type];
			UIDynamicPopup typePopup = CreateScrollablePopup(myPhraseTypeChooser, false);
			myPhraseUI.Add(typePopup);

			UIDynamicButton actionsButton = Utils.SetupButton(this, "Actions " + phrase.Trigger.Name, phrase.Trigger.OpenPanel, false);
			myPhraseUI.Add(actionsButton);

			if (phrase.Type == PHRASE_TYPE_KEYWORD)
			{
				UIDynamicTextInfo keywordHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Phrase Keywords</b></size>", true);
				myPhraseUI.Add(keywordHeader);

				UIDynamicButton addButton = Utils.SetupButton(this, "Add Keyword", UIAddKeyword, true);
				myPhraseUI.Add(addButton);

				int count = phrase.Keywords != null ? phrase.Keywords.Count : 0;
				for (int i=0; i<count; ++i)
				{
					int index = i;
					string keyword = phrase.Keywords[i];
					JSONStorableString keywordStorable = new JSONStorableString("Keyword", keyword, (string kw) => UIChangeKeyword(index, kw));
					UIDynamicInputXButton keywordInput = Utils.SetupInputXButton(this, keywordStorable, () => UIRemoveKeyword(index), true);
					myPhraseUI.Add(keywordInput);
					myPhraseUI.Add(keywordStorable);
				}
			}
			else if (phrase.Type == PHRASE_TYPE_GRAMMAR)
			{
				UIDynamicTextInfo grammarHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Phrase Grammar</b></size>", true);
				myPhraseUI.Add(grammarHeader);

				UIDynamicButton pathButton = CreateButton("Load Grammar XML", true);

				// simple line wrap logic that tries to wrap at slashes, and hard-wraps at maxLine characters
				int maxLine = 28;
				int curLine = 0;
				string path = string.IsNullOrEmpty(phrase.Path) ? string.Empty : phrase.Path;
				StringBuilder builder = new StringBuilder();
				for (int i=0; i<path.Length; ++i)
				{
					builder.Append(path[i]);
					++curLine;
					if ((path[i] == '/' && path.IndexOf('/', i+1, Mathf.Min(maxLine-curLine, path.Length-i-1)) < 0) || (curLine > maxLine))
					{
						curLine = 0;
						builder.Append('\n');
					}
				}

				UIDynamicTextInfo pathInfo = Utils.SetupInfoTextNoScroll(this, builder.ToString(), 200, true);

				myPhrasePath.setCallbackFunction -= UILoadGrammarXML;
				myPhrasePath.allowFullComputerBrowse = false;
				myPhrasePath.allowBrowseAboveSuggestedPath = true;
				if (!string.IsNullOrEmpty(phrase.Path))
					myPhrasePath.SetFilePath(phrase.Path);

				myPhrasePath.RegisterFileBrowseButton(pathButton.button);
				myPhrasePath.setCallbackFunction += UILoadGrammarXML;
				myPhraseUI.Add(pathButton);
				myPhraseUI.Add(pathInfo);
			}
		}

		private void UIAddPhrase()
		{
			UIAddPhrase(null);
		}

		private void UIAddPhrase(Phrase source)
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "Phrase" + i;
				if (!myPhraseChooser.choices.Contains(name))
				{
					Phrase phrase;
					if (source == null)
						phrase = new Phrase(this, name);
					else
						phrase = new Phrase(this, name, source);
					myPhrases.Add(phrase);
					myPhrases.Sort(Phrase.SortByNameAscending);
					UIRebuild();
					myPhraseChooser.val = name;
					return;
				}
			}
			SuperController.LogError("SpeechRecognition: Too many phrases!");
		}

		private void UIDuplicatePhrase()
		{
			UIAddPhrase(UIGetPhrase());
		}

		private void UIRemovePhrase()
		{
			int idx = myPhraseChooser.choices.IndexOf(myPhraseChooser.val);
			if (idx < 0)
				return;

			List<string> options = myPhraseChooser.choices;
			List<string> displays = myPhraseChooser.displayChoices;
			options.RemoveAt(idx);
			displays.RemoveAt(idx);
			myPhrases[idx].Trigger.Remove();
			myPhrases[idx].DisposeRecognizer();
			myPhrases.RemoveAt(idx);
			myPhraseChooser.choices = null; // force UI sync
			myPhraseChooser.choices = options;

			if (idx < options.Count)
				myPhraseChooser.val = options[idx];
			else if (options.Count > 0)
				myPhraseChooser.val = options[options.Count-1];
			else
				myPhraseChooser.val = "";
		}

		private void UITriggerPhrase()
		{
			Phrase phrase = UIGetPhrase();
			if (phrase != null)
				phrase.Trigger.Trigger();
		}

		private void UIRenamePhrase(string name)
		{
			Phrase phrase = UIGetPhrase();
			if (phrase == null || phrase.Name == name)
				return;

			phrase.Name = name;
			myPhrases.Sort(Phrase.SortByNameAscending);
			UIRebuild();
			myPhraseChooser.val = name;
		}

		private void UISetEnabledPhrase(bool v)
		{
			Phrase phrase = UIGetPhrase();
			if (phrase == null)
				return;

			if (phrase.Enabled != v)
			{
				phrase.Enabled = v;
				UIRebuild();
				UIRefresh();
			}
		}

		private void SetEnabledAction(string name)
		{
			mySetEnabled.valNoCallback = string.Empty;
			Phrase phrase = GetPhrase(name);
			if (phrase != null && !phrase.Enabled)
			{
				phrase.Enabled = true;
				myNeedUIRebuild = true;
				myNeedUIRefresh = true;
			}
			else if (phrase == null)
			{
				SuperController.LogError($"SpeechRecognition: Can't enable unknown phrase '{name}'.");
			}
		}

		private void SetDisabledAction(string name)
		{
			mySetDisabled.valNoCallback = string.Empty;
			Phrase phrase = GetPhrase(name);
			if (phrase != null && phrase.Enabled)
			{
				phrase.Enabled = false;
				myNeedUIRebuild = true;
				myNeedUIRefresh = true;
			}
			else if (phrase == null)
			{
				SuperController.LogError($"SpeechRecognition: Can't disable unknown phrase '{name}'.");
			}
		}

		private void SetEnabledAllAction(bool allEnabled)
		{
			for (int i=0; i<myPhrases.Count; ++i)
				myPhrases[i].Enabled = allEnabled;
			myNeedUIRebuild = true;
			myNeedUIRefresh = true;
		}


		private void UISetType(string type)
		{
			Phrase phrase = UIGetPhrase();
			if (phrase == null)
				return;

			int typeIdx = myPhraseTypeChooser.choices.IndexOf(type);
			if (phrase.Type == typeIdx)
				return;
			else if (typeIdx == PHRASE_TYPE_KEYWORD)
				phrase.Keywords = null;
			else if (typeIdx == PHRASE_TYPE_GRAMMAR)
				phrase.Path = null;

			UIRefresh();
		}

		private void UIAddKeyword()
		{
			Phrase phrase = UIGetPhrase();
			if (phrase == null)
				return;

			List<string> keywords = phrase.Keywords;
			if (keywords == null)
				keywords = new List<string>();
			keywords.Add("Keyword!");
			phrase.Keywords = keywords;
			UIRefresh();
		}

		private void UIChangeKeyword(int index, string v)
		{
			Phrase phrase = UIGetPhrase();
			if (phrase == null)
				return;

			List<string> keywords = phrase.Keywords;
			if (keywords != null && index >= 0 && index < keywords.Count)
			{
				keywords[index] = v;
				phrase.Keywords = keywords;
				UIRefresh();
			}
		}

		private void UIRemoveKeyword(int index)
		{
			Phrase phrase = UIGetPhrase();
			if (phrase == null)
				return;

			List<string> keywords = phrase.Keywords;
			if (keywords != null && index >= 0 && index < keywords.Count)
			{
				keywords.RemoveAt(index);
				phrase.Keywords = keywords;
				UIRefresh();
			}
		}

		private void UILoadGrammarXML(string url)
		{
			Phrase phrase = UIGetPhrase();
			if (phrase != null)
			{
				phrase.Path = url;
				UIRefresh();
			}
		}

		private Phrase UIGetPhrase()
		{
			return GetPhrase(myPhraseChooser.val);
		}

		private Phrase GetPhrase(string name)
		{
			int idx = myPhraseChooser.choices.IndexOf(name);
			return idx >= 0 ? myPhrases[idx] : null;
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myPhrases.Count; ++i)
				myPhrases[i].Trigger.SyncAtomNames();
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray plist = new JSONArray();
				for (int i=0; i<myPhrases.Count; ++i)
				{
					Phrase phrase = myPhrases[i];
					JSONClass pc = new JSONClass();
					pc["Name"] = phrase.Name;
					pc["Enabled"].AsBool = phrase.Enabled;

					if (phrase.Path != null)
					{
						pc["Path"] = phrase.Path;
					}
					else if (phrase.Keywords != null && phrase.Keywords.Count > 0)
					{
						JSONArray klist = new JSONArray();
						for (int k=0; k<phrase.Keywords.Count; ++k)
							klist.Add("", phrase.Keywords[k]);
						pc["Keywords"] = klist;
					}
					pc[phrase.Trigger.Name] = phrase.Trigger.GetJSON(base.subScenePrefix);
					plist.Add("", pc);
				}
				jc["Phrases"] = plist;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myPhrases.Count; ++i)
				{
					myPhrases[i].Trigger.Remove();
					myPhrases[i].DisposeRecognizer();
				}
				myPhrases.Clear();

				JSONArray clist = jc["Phrases"].AsArray;
				for (int i=0; i<clist.Count; ++i)
				{
					JSONClass pc = clist[i].AsObject;
					Phrase phrase = new Phrase(this, pc["Name"]);
					phrase.Enabled = pc["Enabled"].AsBool;

					if (pc.HasKey("Path"))
					{
						phrase.Path = pc["Path"];
					}
					else if (pc.HasKey("Keywords"))
					{
						JSONArray klist = pc["Keywords"].AsArray;
						List<string> keywords = new List<string>(klist.Count);
						for (int k=0; k<klist.Count; ++k)
							keywords.Add(klist[k].Value);
						phrase.Keywords = keywords;
					}

					phrase.Trigger.RestoreFromJSON(pc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myPhrases.Add(phrase);
				}

				myPhrases.Sort(Phrase.SortByNameAscending);
				UIRebuild();

				List<string> options = myPhraseChooser.choices;
				if (options.Count == 0)
					myPhraseChooser.valNoCallback = "";
				else if (!options.Contains(myPhraseChooser.val))
					myPhraseChooser.valNoCallback = options[0];

				myNeedUIRefresh = true;
			}
		}

		private void Update()
		{
			for (int i=0; i<myPhrases.Count; ++i)
			{
				myPhrases[i].Trigger.Update();
			}

			if (UITransform.gameObject.activeInHierarchy)
			{
				if (myNeedUIRebuild)
					UIRebuild();
				if (myNeedUIRefresh)
					UIRefresh();
			}
		}

		private void UIRefresh()
		{
			myNeedUIRefresh = false;
			UISetPhrase(myPhraseChooser.val);
		}

		private void UIRebuild()
		{
			myNeedUIRebuild = false;
			List<string> options = myPhraseChooser.choices;
			List<string> displays = myPhraseChooser.displayChoices;
			options.Clear();
			displays.Clear();
			for (int i=0; i<myPhrases.Count; ++i)
			{
				Phrase phrase = myPhrases[i];
				options.Add(phrase.Name);
				displays.Add(phrase.Enabled ? phrase.Name : $"<color=#CC0000>{phrase.Name}</color>");
			}
			myPhraseChooser.choices = null; // force UI sync
			myPhraseChooser.choices = options;
			string current = myPhraseChooser.val;
			myPhraseChooser.valNoCallback = "";
			myPhraseChooser.valNoCallback = current;
		}

		private void OnEnable()
		{
			for (int i=0; i<myPhrases.Count; ++i)
				myPhrases[i].StartRecognizer();
		}

		private void OnDisable()
		{
			for (int i=0; i<myPhrases.Count; ++i)
				myPhrases[i].StopRecognizer();
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			for (int i=0; i<myPhrases.Count; ++i)
			{
				myPhrases[i].Trigger.Remove();
				myPhrases[i].DisposeRecognizer();
			}
			myPhrases.Clear();

			Utils.OnDestroyUI();

			PhraseRecognitionSystem.OnError -= OnPhaseSystemError;
		}

		private void OnPhaseSystemError(SpeechError errorCode)
		{
			SuperController.LogError("SpeechRecognition: " + errorCode);
		}

		private static void ExtractFilesFromPackage(string path)
		{
			string[] files = FileManagerSecure.GetFiles(path);
			foreach (string pkgfile in files)
			{
				if (!pkgfile.EndsWith("." + FILE_EXTENSION))
					continue;
				SuperController.LogMessage("Extracting file from VAR: " + pkgfile);
				string newfile = EXTRACT_PATH + pkgfile.Replace(":\\", "\\");
				string newdir = FileManagerSecure.GetDirectoryName(newfile);
				if (!FileManagerSecure.DirectoryExists(newdir))
					FileManagerSecure.CreateDirectory(newdir);
				FileManagerSecure.CopyFile(pkgfile, newfile);
			}

			string[] dirs = FileManagerSecure.GetDirectories(path);
			foreach (string dir in dirs)
			{
				ExtractFilesFromPackage(dir);
			}
		}
    }
}
